# frozen_string_literal: true

module FrozenRecord
  module Backends
    module Yaml
      autoload :ERB, 'erb'

      extend self

      # Transforms the model name into a valid filename.
      #
      # @param format [String] the model name that inherits
      #   from FrozenRecord::Base
      # @return [String] the file name.
      def filename(model_name)
        "#{model_name.underscore.pluralize}.yml"
      end

      # Reads file in `file_path` and return records.
      #
      # @param format [String] the file path
      # @return [Array] an Array of Hash objects with keys being attributes.
      def load(file_path)
        if !File.exist?(file_path) && File.exist?("#{file_path}.erb")
          file_path = "#{file_path}.erb"
        end

        if FrozenRecord.deprecated_yaml_erb_backend
          yml_erb_data = File.read(file_path)
          yml_data = ERB.new(yml_erb_data).result

          unless file_path.end_with?('.erb')
            if yml_data != yml_erb_data
              basename = File.basename(file_path)
              raise "[FrozenRecord] Deprecated: `#{basename}` contains ERB tags and should be renamed `#{basename}.erb`.\nSet FrozenRecord.deprecated_yaml_erb_backend = false to enable the future behavior"
            end
          end

          load_string(yml_data)
        else
          if file_path.end_with?('.erb')
            load_string(ERB.new(File.read(file_path)).result)
          else
            load_file(file_path)
          end
        end
      end

      private

      supports_freeze = begin
        YAML.load_file(File.expand_path('../empty.yml', __FILE__), freeze: true)
      rescue ArgumentError
        false
      end

      if YAML.respond_to?(:unsafe_load_file)
        def load_file(path)
          YAML.unsafe_load_file(path, freeze: true) || [].freeze
        end

        def load_string(yaml)
          YAML.unsafe_load(yaml, freeze: true) || [].freeze
        end
      elsif supports_freeze
        def load_file(path)
          YAML.load_file(path, freeze: true) || [].freeze
        end

        def load_string(yaml)
          YAML.load(yaml, freeze: true) || [].freeze
        end
      else
        def load_file(path)
          YAML.load_file(path) || [].freeze
        end

        def load_string(yaml)
          YAML.load(yaml) || [].freeze
        end
      end
    end
  end
end
